==> ssh-2.2.0.tar.gz: ssh-2.2.0/lib/sshutil/sshstrutil/sshbitvector.c <==
/* 

  sshbitvector.c

  Author: Jukka Aittokallio <jai@ssh.fi>

  Copyright (c) 2000 SSH Communications Security, Finland
                   All rights reserved

  This file contains functions for creating and 
  manipulating bit vectors. Bit vector is an arbitrary
  sized "boolean" array. You can set/get/query bit status on
  the array.
    
*/

#include "sshbitvector.h"

struct SshBitVectorRec {

  Boolean fixed_size;     /* is vector fixed_size */
  SshUInt32 size,         /* allocated size in bytes */
            bit_count;    /* current size in bits */
  unsigned char *bytes;   /* bit data in byte array. 8 bits per byte */

};

SshBitVector ssh_bitvector_create(SshUInt32 initial_size)
{
  int i;
  SshBitVector v;

  if (!(v = (SshBitVector)ssh_xmalloc(sizeof(*v))))
    {
      return NULL;
    }

  if (initial_size > 0)
    {
      v->fixed_size = 1;
      v->bit_count = initial_size;
      v->size = (initial_size + 7) / 8;
      v->bytes = (unsigned char *)ssh_xmalloc(v->size);
      if (!v->bytes)
        {
          ssh_xfree(v);
          return NULL;
        }
      for (i=0; i<v->size; i++) v->bytes[i] = 0;
    }
  else
    {
      v->bit_count = 0;
      v->size = 0;
      v->fixed_size = 0;
      v->bytes = NULL;
    }
  return v;
}

static unsigned char ssh_swap_bits(unsigned char b)
{
  unsigned char r = 0, m1 = 0x01, m2 = 0x80;
  while(m2)
    {
      if (b & m2) r |= m1;    
      m1<<=1;
      m2>>=1;
    }
  return r;
}

SshBitVector ssh_bitvector_import(unsigned char *byte_array, 
                                                                                                                                        SshUInt32 bit_count, 
                                  Boolean fixed_size, 
                                                                                                                                        Boolean swap_bits)
{
  SshUInt32 i;
  SshBitVector v;

  if (!(v = (SshBitVector)ssh_xcalloc(1, sizeof(*v))))
    {
      return NULL;
    }
  v->bit_count = bit_count;
  v->size = (bit_count + 7) / 8;
  v->fixed_size = fixed_size;

        if (v->size)
                {
                        v->bytes = (unsigned char *)ssh_xmalloc(v->size);
                        if (!v->bytes)
                                {
                                        ssh_xfree(v);
                                        return NULL;
                                }
                        for (i=0; i<v->size; i++) 
                                v->bytes[i] = (swap_bits ? 
                                        ssh_swap_bits(byte_array[i]) : byte_array[i]);
                }
  return v;
}

SshBitStatus ssh_bitvector_export(SshBitVector v, 
                                                                                                                                        unsigned char **byte_array, 
                                  SshUInt32 *bit_count)
{
  SshUInt32 i;

  *bit_count = v->bit_count;
  *byte_array = (unsigned char *)ssh_xmalloc(v->size);

  if (!*byte_array)
    {
      return SSH_BITVEC_ALLOC_ERROR;
    }

  for (i=0; i<v->size; i++) (*byte_array)[i] = v->bytes[i];

  return SSH_BITVEC_OK;
}

void ssh_bitvector_destroy(SshBitVector v)
{
  if (v)
    {
      ssh_xfree(v->bytes);
      ssh_xfree(v);
    }
}

SshBitStatus ssh_bitvector_resize(SshBitVector v, SshUInt32 bit_count)
{
  unsigned char *t;
  SshUInt32 size = bit_count / 8 + 1;

  if (v->fixed_size)
    {
      return SSH_BITVEC_FIXED_SIZE;
    }

  if (size != v->size)
    {
      t = (unsigned char *)ssh_xrealloc(v->bytes, size);
      if (!t)
        {
          return SSH_BITVEC_ALLOC_ERROR;
        }
      v->bytes = t;
      v->size = size;
      v->bit_count = bit_count;
    }
  return SSH_BITVEC_OK;
}

/*
 * Internally used function to check bit indexes and also to allocate
 * more space for bits.
 *
 */
static SshBitStatus handle_bit_index(SshBitVector v, 
                                                                                                                                                 SshUInt32 bit_num, 
                                     Boolean can_alloc)
{
  SshUInt32 size;
  unsigned char *t;

  if (bit_num > v->bit_count - 1)
    {
      if (!can_alloc)
        {
          return SSH_BITVEC_INVALID_INDEX;
        }
      size = bit_num / 8 + 1;
      if (size > v->size)
        {
          t = ssh_xrealloc(v->bytes, size);
          if (!t)
            {
              return SSH_BITVEC_ALLOC_ERROR;
            }
          v->bytes = t;
          v->size = size;
          v->bit_count = bit_num;
        }
        
    }
  return SSH_BITVEC_OK;
}

SshBitStatus ssh_bitvector_set_bit(SshBitVector v, SshUInt32 bit_num)
{
  SshBitStatus r = handle_bit_index(v, bit_num, !v->fixed_size);
  if (r != SSH_BITVEC_OK)
    {
      return r;
    }
  v->bytes[bit_num / 8] |= (1 << (bit_num % 8));
  return SSH_BITVEC_OK;    
}


SshBitStatus ssh_bitvector_clear_bit(SshBitVector v, SshUInt32 bit_num)
{
  SshBitStatus r = handle_bit_index(v, bit_num, 0);

  if (r != SSH_BITVEC_OK)
    {
      if (v->fixed_size)
        return SSH_BITVEC_INVALID_INDEX;
    }
  else
    {
      v->bytes[bit_num / 8] &= ~(1 << (bit_num % 8));
    }
  return SSH_BITVEC_OK;    
}

SshBitStatus ssh_bitvector_query_bit(SshBitVector v, SshUInt32 bit_num)
{
  SshBitStatus r = handle_bit_index(v, bit_num, 0);

  if (r != SSH_BITVEC_OK)
    {
      if (v->fixed_size)
        return SSH_BITVEC_INVALID_INDEX;
      else
        return SSH_BITVEC_BIT_OFF;
    }
  else
    {
      if (v->bytes[bit_num / 8] & (1 << (bit_num % 8)))
        return SSH_BITVEC_BIT_ON;
      else
        return SSH_BITVEC_BIT_OFF;
    }
}

static int find_clear_bit_in_byte(unsigned char b)
{
  unsigned char mask = 0x01;
  int p = 0;

  if (b == 0) return 0;
  do
    { 
      if (!(b & mask))
        return p;
      mask<<=1;
      p++;
    }
  while (mask);
  return -1;
}

static void dump_byte_bits(unsigned char b, int *bits_left)
{
  unsigned char mask = 0x01;
  char s[9];
  int p = 0;

  do
    {
      if (b & mask)
        s[p++] = '1';
      else
        {
           s[p++] = '0';
        }
      (*bits_left)--;
      mask<<=1;
    }      
  while (mask && *bits_left);
  s[p] = '\0';
  printf("%s", s);
}

SshBitStatus ssh_bitvector_dump(SshBitVector v)
{
  int i, bits_left = v->bit_count;

  for (i=0; i<v->size; i++)
    dump_byte_bits(v->bytes[i], &bits_left);
  return SSH_BITVEC_OK;
}

SshBitStatus ssh_bitvector_find_first_clear_bit(SshBitVector v, SshUInt32 *bit_pos)
{
  SshUInt32 i;
  int p;

  for (i=0; i<v->size; i++)
    {
      p = find_clear_bit_in_byte(v->bytes[i]);
      if (p != -1) 
        {
          *bit_pos = i*8+p;
          return SSH_BITVEC_OK;
        }
    }
  if (v->fixed_size)
    return SSH_BITVEC_NOT_FOUND;
  *bit_pos = v->bit_count;
  return SSH_BITVEC_OK;  
}

SshUInt32 ssh_bitvector_get_bit_count(SshBitVector v)
{
  return v->bit_count;
}

==> ssh-2.4.0.tar.gz: ssh-2.4.0/lib/sshutil/sshstrutil/sshbitvector.c <==
/* 

  sshbitvector.c

  Author: Jukka Aittokallio <jai@ssh.fi>

  Copyright (c) 2000 SSH Communications Security, Finland
                   All rights reserved

  This file contains functions for creating and 
  manipulating bit vectors. Bit vector is an arbitrary
  sized "boolean" array. You can set/get/query bit status on
  the array.
    
*/

#include "sshbitvector.h"

struct SshBitVectorRec {

  Boolean fixed_size;     /* is vector fixed_size */
  SshUInt32 size,         /* allocated size in bytes */
            bit_count;    /* current size in bits */
  unsigned char *bytes;   /* bit data in byte array. 8 bits per byte */

};

SshBitVector ssh_bitvector_create(SshUInt32 initial_size)
{
  int i;
  SshBitVector v;

  if (!(v = (SshBitVector)ssh_xmalloc(sizeof(*v))))
    {
      return NULL;
    }

  if (initial_size > 0)
    {
      v->fixed_size = 1;
      v->bit_count = initial_size;
      v->size = (initial_size + 7) / 8;
      v->bytes = (unsigned char *)ssh_xmalloc(v->size);
      if (!v->bytes)
        {
          ssh_xfree(v);
          return NULL;
        }
      for (i=0; i<v->size; i++) v->bytes[i] = 0;
    }
  else
    {
      v->bit_count = 0;
      v->size = 0;
      v->fixed_size = 0;
      v->bytes = NULL;
    }
  return v;
}

static const unsigned char swap_table[256] = {
  0x00, 0x80, 0x40, 0xC0, 0x20, 0xA0, 0x60, 0xE0,
  0x10, 0x90, 0x50, 0xD0, 0x30, 0xB0, 0x70, 0xF0,
  0x08, 0x88, 0x48, 0xC8, 0x28, 0xA8, 0x68, 0xE8,
  0x18, 0x98, 0x58, 0xD8, 0x38, 0xB8, 0x78, 0xF8,
  0x04, 0x84, 0x44, 0xC4, 0x24, 0xA4, 0x64, 0xE4,
  0x14, 0x94, 0x54, 0xD4, 0x34, 0xB4, 0x74, 0xF4,
  0x0C, 0x8C, 0x4C, 0xCC, 0x2C, 0xAC, 0x6C, 0xEC,
  0x1C, 0x9C, 0x5C, 0xDC, 0x3C, 0xBC, 0x7C, 0xFC,
  0x02, 0x82, 0x42, 0xC2, 0x22, 0xA2, 0x62, 0xE2,
  0x12, 0x92, 0x52, 0xD2, 0x32, 0xB2, 0x72, 0xF2,
  0x0A, 0x8A, 0x4A, 0xCA, 0x2A, 0xAA, 0x6A, 0xEA,
  0x1A, 0x9A, 0x5A, 0xDA, 0x3A, 0xBA, 0x7A, 0xFA,
  0x06, 0x86, 0x46, 0xC6, 0x26, 0xA6, 0x66, 0xE6,
  0x16, 0x96, 0x56, 0xD6, 0x36, 0xB6, 0x76, 0xF6,
  0x0E, 0x8E, 0x4E, 0xCE, 0x2E, 0xAE, 0x6E, 0xEE,
  0x1E, 0x9E, 0x5E, 0xDE, 0x3E, 0xBE, 0x7E, 0xFE,
  0x01, 0x81, 0x41, 0xC1, 0x21, 0xA1, 0x61, 0xE1,
  0x11, 0x91, 0x51, 0xD1, 0x31, 0xB1, 0x71, 0xF1,
  0x09, 0x89, 0x49, 0xC9, 0x29, 0xA9, 0x69, 0xE9,
  0x19, 0x99, 0x59, 0xD9, 0x39, 0xB9, 0x79, 0xF9,
  0x05, 0x85, 0x45, 0xC5, 0x25, 0xA5, 0x65, 0xE5,
  0x15, 0x95, 0x55, 0xD5, 0x35, 0xB5, 0x75, 0xF5,
  0x0D, 0x8D, 0x4D, 0xCD, 0x2D, 0xAD, 0x6D, 0xED,
  0x1D, 0x9D, 0x5D, 0xDD, 0x3D, 0xBD, 0x7D, 0xFD,
  0x03, 0x83, 0x43, 0xC3, 0x23, 0xA3, 0x63, 0xE3,
  0x13, 0x93, 0x53, 0xD3, 0x33, 0xB3, 0x73, 0xF3,
  0x0B, 0x8B, 0x4B, 0xCB, 0x2B, 0xAB, 0x6B, 0xEB,
  0x1B, 0x9B, 0x5B, 0xDB, 0x3B, 0xBB, 0x7B, 0xFB,
  0x07, 0x87, 0x47, 0xC7, 0x27, 0xA7, 0x67, 0xE7,
  0x17, 0x97, 0x57, 0xD7, 0x37, 0xB7, 0x77, 0xF7,
  0x0F, 0x8F, 0x4F, 0xCF, 0x2F, 0xAF, 0x6F, 0xEF,
  0x1F, 0x9F, 0x5F, 0xDF, 0x3F, 0xBF, 0x7F, 0xFF
};

SshBitVector ssh_bitvector_import(unsigned char *byte_array, 
                                  SshUInt32 bit_count, 
                                  Boolean fixed_size, 
                                  Boolean swap_bits)
{
  SshUInt32 i;
  SshBitVector v;

  if (!(v = (SshBitVector)ssh_xcalloc(1, sizeof(*v))))
    {
      return NULL;
    }
  v->bit_count = bit_count;
  v->size = (bit_count + 7) / 8;
  v->fixed_size = fixed_size;

  if (v->size)
    {
      v->bytes = (unsigned char *)ssh_xmalloc(v->size);
      if (!v->bytes)
        {
          ssh_xfree(v);
          return NULL;
        }
      for (i=0; i<v->size; i++) 
        v->bytes[i] = ((swap_bits)? 
                       (swap_table[byte_array[i]]):(byte_array[i]));
    }
  return v;
}

SshBitStatus ssh_bitvector_export(SshBitVector v,
                                  unsigned char **byte_array, 
                                  SshUInt32 *bit_count,
                                  Boolean swap_bits)
{
  SshUInt32 i;

  *bit_count = v->bit_count;
  *byte_array = (unsigned char *)ssh_xmalloc(v->size);

  if (!*byte_array)
    {
      return SSH_BITVEC_ALLOC_ERROR;
    }

  for (i=0; i<v->size; i++) 
    (*byte_array)[i] = ((swap_bits)?(swap_table[v->bytes[i]]):(v->bytes[i]));

  return SSH_BITVEC_OK;
}

void ssh_bitvector_destroy(SshBitVector v)
{
  if (v)
    {
      ssh_xfree(v->bytes);
      ssh_xfree(v);
    }
}

SshBitStatus ssh_bitvector_resize(SshBitVector v, SshUInt32 bit_count)
{
  unsigned char *t;
  SshUInt32 size = bit_count / 8 + 1;

  if (size != v->size)
    {
      t = (unsigned char *)ssh_xrealloc(v->bytes, size);
      if (!t)
        {
          return SSH_BITVEC_ALLOC_ERROR;
        }
      v->bytes = t;
      if (size > v->size)
        memset(&v->bytes[v->size], 0, size - v->size);  
      v->size = size;
      v->bit_count = bit_count;
    }
  return SSH_BITVEC_OK;
}

/*
 * Internally used function to check bit indexes and also to allocate
 * more space for bits.
 *
 */
static SshBitStatus handle_bit_index(SshBitVector v, 
                                     SshUInt32 bit_num, 
                                     Boolean can_alloc)
{
  SshUInt32 size;
  unsigned char *t;

  if (bit_num + 1 > v->bit_count)
    {
      if (!can_alloc)
        {
          return SSH_BITVEC_INVALID_INDEX;
        }
      size = bit_num / 8 + 1;
      if (size > v->size)
        {
          t = ssh_xrealloc(v->bytes, size);
          if (!t)
            {
              return SSH_BITVEC_ALLOC_ERROR;
            }
          v->bytes = t;
          memset(&v->bytes[v->size], 0, size - v->size);
          v->size = size;
        }
      v->bit_count = bit_num + 1;        
    }
  return SSH_BITVEC_OK;
}

SshBitStatus ssh_bitvector_set_bit(SshBitVector v, SshUInt32 bit_num)
{
  SshBitStatus r = handle_bit_index(v, bit_num, !v->fixed_size);
  if (r != SSH_BITVEC_OK)
    {
      return r;
    }
  v->bytes[bit_num / 8] |= (1 << (bit_num % 8));
  return SSH_BITVEC_OK;    
}


SshBitStatus ssh_bitvector_clear_bit(SshBitVector v, SshUInt32 bit_num)
{
  SshBitStatus r = handle_bit_index(v, bit_num, 0);

  if (r != SSH_BITVEC_OK)
    {
      if (v->fixed_size)
        return SSH_BITVEC_INVALID_INDEX;
    }
  else
    {
      v->bytes[bit_num / 8] &= ~(1 << (bit_num % 8));
    }
  return SSH_BITVEC_OK;    
}

SshBitStatus ssh_bitvector_query_bit(SshBitVector v, SshUInt32 bit_num)
{
  SshBitStatus r = handle_bit_index(v, bit_num, 0);

  if (r != SSH_BITVEC_OK)
    {
      if (v->fixed_size)
        return SSH_BITVEC_INVALID_INDEX;
      else
        return SSH_BITVEC_BIT_OFF;
    }
  else
    {
      if (v->bytes[bit_num / 8] & (1 << (bit_num % 8)))
        return SSH_BITVEC_BIT_ON;
      else
        return SSH_BITVEC_BIT_OFF;
    }
}

static int find_clear_bit_in_byte(unsigned char b)
{
  unsigned char mask = 0x01;
  int p = 0;

  if (b == 0) return 0;
  do
    { 
      if (!(b & mask))
        return p;
      mask<<=1;
      p++;
    }
  while (mask);
  return -1;
}

static void dump_byte_bits(unsigned char b, int *bits_left)
{
  unsigned char mask = 0x01;
  char s[9];
  int p = 0;

  do
    {
      if (b & mask)
        s[p++] = '1';
      else
        {
           s[p++] = '0';
        }
      (*bits_left)--;
      mask<<=1;
    }      
  while (mask && *bits_left);
  s[p] = '\0';
  printf("%s", s);
}

SshBitStatus ssh_bitvector_dump(SshBitVector v)
{
  int i, bits_left = v->bit_count;

  for (i=0; i<v->size; i++)
    dump_byte_bits(v->bytes[i], &bits_left);
  return SSH_BITVEC_OK;
}

SshBitStatus ssh_bitvector_find_first_clear_bit(SshBitVector v, SshUInt32 *bit_pos)
{
  SshUInt32 i;
  int p;

  for (i=0; i<v->size; i++)
    {
      p = find_clear_bit_in_byte(v->bytes[i]);
      if (p != -1) 
        {
          *bit_pos = i*8+p;
          return SSH_BITVEC_OK;
        }
    }
  if (v->fixed_size)
    return SSH_BITVEC_NOT_FOUND;
  *bit_pos = v->bit_count;
  return SSH_BITVEC_OK;  
}

SshUInt32 ssh_bitvector_get_bit_count(SshBitVector v)
{
  return v->bit_count;
}
